# coding: utf-8

from tomwer.tests.conftest import qtapp  # noqa F401
from orangecontrib.tomwer.widgets.control.AdvancementOW import AdvancementOW


def test_totsaso(qtapp):  # noqa F401
    AdvancementOW()
