"""
Contains tools used by the orange add-on and used to keep the orange-add-on and
gui 'alive', avoid gui freeze during processing.

With time it should be replace by the ewoksorange mechanism (inheritance to ewoksorange widgets)
"""
