# import pickle

# from orangecanvas.scheme.readwrite import literal_dumps

# from tomwer.tests.conftest import qtapp  # noqa F401

# from orangecontrib.tomwer.widgets.reconstruction.NabuHelicalPrepareWeightsDoubleOW import (
#     NabuHelicalPrepareWeightsDoubleOW,
# )


# def test_NabuHelicalPrepareWeightsDoubleOW(
#     qtapp,  # noqa F811
# ):
#     """simple test of the _DeltaBetaSelectorDialog"""
#     widget = NabuHelicalPrepareWeightsDoubleOW()
#     widget._loadSettings()
#     # test settings serialization
#     pickle.dumps(widget._ewoks_default_inputs)
#     literal_dumps(widget._ewoks_default_inputs)
